import click

from operator import itemgetter
from switchbladecli.cli.bundle_cache import Bundle


@click.command()
@click.pass_context
def update(ctx):
    project_dir, config, verbose = itemgetter("project_dir", "config", "verbose")(ctx.obj)
    cmd_update(verbose, project_dir, config)


def cmd_update(verbose: bool, project_dir: str, config: dict):
    # Instantiating a Bundle object will fetch the latest bundle from the source or cache
    Bundle(config)



import click

from operator import itemgetter
from switchbladecli.modes.tool_runner import get_tool_runner


@click.command()
@click.argument("test", required=False)
@click.pass_context
def test(ctx, test: str):
    """Run testing tool(s) on project.
    
    TEST: Testing tool to run. If not specified, all testing tools will be run.
    """
    project_dir, config, verbose = itemgetter("project_dir", "config", "verbose")(ctx.obj)
    result = cmd_test(verbose, project_dir, config, test)
    if not result:
        ctx.exit(1)


def cmd_test(verbose: bool, project_dir: str, config: dict, test_tool: str = "all"):
    tool_runner = get_tool_runner(config)(config, verbose)
    return tool_runner.test(test_tool)



import click
import os
import shutil

from datetime import datetime, timezone
from github import Github
from mergedeep import merge
from pathlib import Path
from tomlkit import dumps, loads

from switchbladecli.utils import hash_dir

BUNDLE_CONFIG_FILES = ["bundle.toml"]
CACHE_FOLDERNAME = ".switchblade-cache"
LATEST_CONFIG_FILENAME = "latest.toml"
UPDATE_LOG_FILENAME = "update.log"


class Bundle:
    """A bundle of dev tools and configs.
    """

    def __init__(self, switchblade_config: dict, version: str = None):
        """Initialize a Bundle object.

        If version is None, the latest bundle will be used (conditionally fetched and cached).

        If a version is passed:
            1. Tries to fetch a bundle with that version from the cache.
            2. If the bundle doesn't exist, an exception is raised.

        Args:
            switchblade_config (dict): The switchblade config.
            version (str, optional): The version of the bundle to fetch. Defaults to None.
        """
        self._switchblade_config = switchblade_config
        if version is None:
            click.echo("⚔️ Switchblade initializing bundle...")
            cache = BundleCache(self._switchblade_config)
            bundle = self._fetch_latest(cache)
            self.version = bundle.version
            self.bundle_folder = bundle.bundle_folder
            self.bundle_config = bundle.bundle_config

        else:
            click.echo(f"⚔️ Switchblade initializing bundle with version {version}...")
            self.version = version
            self.bundle_folder = self._bundle_folder_from_version(version)
            self.bundle_config = self._load_bundle_config()
    

    def _load_bundle_config(self) -> dict:
        """Get the bundle config from the config file."""
        bundle_config_file = None
        for bundle_config_file_name in BUNDLE_CONFIG_FILES:
            bundle_config_file = self.bundle_folder / bundle_config_file_name
            if bundle_config_file.exists():
                break
        if bundle_config_file is None or not bundle_config_file.exists():
            raise FileNotFoundError(f"No bundle config file found in bundle {self.version}")
        return loads(bundle_config_file.read_text(encoding="utf-8"))
    
    
    def _bundle_folder_from_version(self, version) -> Path:
        """Get the bundle folder."""
        return Path(CACHE_FOLDERNAME) / version


    def get_files(self) -> list:
        """Get the list of files (excluding the config) in the bundle.
        
        Returns a list of file paths relative to the bundle folder.
        """
        return [str(file.relative_to(self.bundle_folder)) for file in self.bundle_folder.rglob("*") if file.name not in BUNDLE_CONFIG_FILES]


    def _fetch_latest(self, cache: "BundleCache") -> "Bundle":
        """Conditionally fetch and return the latest bundle.
        """
        bundle_source_uri = self._switchblade_config["switchblade"]["bundle"]
        click.echo("⚔️ Switchblade checking for updates...")

        # Check the remote version to see if we need to update
        remote_version = None
        pygithub = None
        if bundle_source_uri.startswith("gh:"):
            try:
                # Init the Github API
                pygithub = Github(os.environ.get("GITHUB_TOKEN"))
                # Get the org name, repo name and folder(s)
                org_name, repo_name, *folder = bundle_source_uri[3:].split("/")
                # Get the commit SHA from the repo (default branch)
                repo = pygithub.get_repo(f"{org_name}/{repo_name}")
                remote_version = repo.get_branch(repo.default_branch).commit.sha
            except Exception as exc:
                # If something went wrong fetching from Github,
                # we use the latest cached bundle if it exists
                latest_cached_version = cache.get_latest_version()
                if latest_cached_version is not None:
                    click.echo(f"⚔️ Switchblade failed to fetch remote bundle, using cached version {latest_cached_version} instead.")
                    cache.log(f"FETCH FAILED_WITH_FALLBACK")
                    return Bundle(self._switchblade_config, latest_cached_version)
                else:
                    # If there is no cached bundle, raise the exception
                    cache.log(f"FETCH FAILED")
                    raise click.ClickException(f"Remote bundle repo '{bundle_source_uri}' could not be fetched: {exc}")
        else:
            # If the folder doesn't exist, raise an exception
            bundle_source_folder = Path(bundle_source_uri)
            if not bundle_source_folder.exists():
                raise FileNotFoundError(f"Bundle folder {bundle_source_uri} does not exist.")
            remote_version = hash_dir(bundle_source_uri)

        cache.log(f"UPDATING {remote_version}")
        
        # Check if the latest version from the remote is already cached
        # (does not necessarily have to be the most recently fetched bundle from 'latest.toml')
        if cache.has_bundle(remote_version):
            # If the latest cached bundle is up to date, just return it
            click.echo(f"⚔️ Switchblade cache already contains version {remote_version}")
            cache.log(f"UPDATING {remote_version} SKIPPED_ALREADY_CACHED")
            return Bundle(self._switchblade_config, remote_version)

        # If the cache is stale, update it
        click.echo(f"⚔️ Switchblade updating tool bundle to version {remote_version}...")
        bundle_folder = self._bundle_folder_from_version(remote_version)
        if bundle_source_uri.startswith("gh:"):
            # Create the bundle folder
            bundle_folder.mkdir(parents=True)
            # Get the contents of the repo
            org_name, repo_name, *folder = bundle_source_uri[3:].split("/")
            repo = pygithub.get_repo(f"{org_name}/{repo_name}")
            repo_contents = repo.get_contents("/".join(folder), ref=remote_version)
            # Download all the files
            for repo_asset in repo_contents:
                with open(bundle_folder / repo_asset.name, "wb") as file:
                    file.write(repo_asset.decoded_content)

        else:
            # If the source is a local folder, copy it to the cache
            shutil.copytree(str(bundle_source_uri), str(bundle_folder), dirs_exist_ok=True)

        cache.log(f"UPDATING {remote_version} SUCCEEDED")

        new_bundle = Bundle(self._switchblade_config, remote_version)
        cache.set_latest_config(new_bundle)

        # Finally, return the fetched bundle
        return new_bundle


class BundleCache:

    def __init__(self, switchblade_config: dict):
        self._switchblade_config = switchblade_config
        self._project_folder = Path(self._switchblade_config.get("project_dir"))
        self._cache_folder = self._initialize_cache_folder(self._project_folder)


    def _initialize_cache_folder(self, project_folder: Path):
        """Initialize a new cache folder if it doesn't exist yet.

        Creates:
            - The cache folder
            - A .gitignore file in the cache folder with an asterisk in it
            - An update.log file in the cache folder
        """
        cache_folder = project_folder / CACHE_FOLDERNAME
        if not cache_folder.exists():
            # Make sure the cache folder exists
            cache_folder.mkdir()
            # Add a .gitignore file with an asterisk in it to ensure that the
            # cache folder is never committed to git
            with open(cache_folder / ".gitignore", "w") as gitignore:
                gitignore.write("*")

        # Make sure the update.log file exists
        update_log = cache_folder / UPDATE_LOG_FILENAME
        if not update_log.exists():
            update_log.touch()
        
        return cache_folder


    def get_bundle(self, version: str) -> Bundle:
        """Get a bundle from the cache."""
        return Bundle(self._switchblade_config, version)


    def get_bundles(self) -> list:
        """Get all bundles in the cache."""
        bundles = []
        for bundle_folder in os.listdir(self._cache_folder):
            # Skip non-folders
            if not (self._cache_folder / bundle_folder).is_dir():
                continue
            bundles.append(Bundle(self._switchblade_config, bundle_folder))
        return bundles


    def _get_latest_config_file(self, check_exists: bool = True) -> Path:
        """Get the latest config file."""
        latest_config_file = self._cache_folder / LATEST_CONFIG_FILENAME
        if check_exists and not latest_config_file.exists():
            return None
        return latest_config_file


    def get_latest_config(self) -> dict:
        """Get the latest config."""
        latest_config_file = self._get_latest_config_file()
        # If the latest.toml file doesn't exist, return None
        if latest_config_file is None:
            return None
        latest = loads(latest_config_file.read_text(encoding="utf-8"))
        return latest["latest"]


    def set_latest_config(self, bundle: Bundle):
        """Set the latest config."""
        latest_config_file = self._get_latest_config_file(False)
        with open(latest_config_file, "w") as latest_toml:
            latest_toml.write(
                dumps(
                    {
                        "latest": {
                            "commit_sha": bundle.version,
                            "fetched_on": datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
                            "files": bundle.get_files(),
                            "last_installed_on": ""
                        }
                    }
                )
            )


    def update_latest_config(self, updated_latest: dict):
        """Update the latest config, merging with updated dict."""
        latest_config_file = self._get_latest_config_file(False)
        latest_config = self.get_latest_config()
        merge(latest_config, updated_latest)
        with open(latest_config_file, "w") as latest_toml:
            latest_toml.write(
                dumps(
                    {
                        "latest": latest_config
                    }
                )
            )


    def get_latest_version(self) -> str:
        """Get the latest version."""
        latest_config = self.get_latest_config()
        if latest_config is None:
            return None
        return latest_config["commit_sha"]


    def get_latest_bundle(self) -> Bundle:
        """Get the latest bundle."""
        latest_config = self.get_latest_config()
        if latest_config is None:
            return None
        return self.get_bundle(latest_config["commit_sha"])


    def log(self, message):
        current_time = datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        with open(self._cache_folder / UPDATE_LOG_FILENAME, "a") as update_log:
            update_log.write(f"[{current_time}] {message}\n")


    def has_bundle(self, version: str) -> bool:
        """Check if a version is in the cache."""
        return (self._cache_folder / version).exists()


    def cleanup(self):
        """Remove all bundles from cache except the latest.
        """
        latest_version = self.get_latest_version()
        bundles = self.get_bundles()
        for bundle in bundles:
            if bundle.version != latest_version:
                shutil.rmtree(bundle.bundle_folder)



import os
from pathlib import Path

import click
from tomlkit import loads

from switchbladecli.exceptions import MissingConfiguration, MissingRequiredConfigKey

CONFIG_FILE_NAMES = [".switchblade", "switchblade.toml"]
REQUIRED_CONFIG_KEYS = {
    "switchblade": ["mode", "bundle"]
}

def find_config_file(verbose: bool, config_file: str) -> str:
    # If no config file is specified, locate it automatically
    if config_file is None:
        for name in CONFIG_FILE_NAMES:
            if os.path.exists(name):
                config_file = name
                break

    if config_file is None or not os.path.exists(config_file):
        raise MissingConfiguration(f"Configuration file not found. Allowed config file names: {CONFIG_FILE_NAMES}")

    return config_file


def get_switchblade_config(verbose, project_dir: str, config_file: str):
    config_file = os.path.join(project_dir, config_file)
    if verbose:
        click.echo(f"Loading configuration from {config_file}")

    config = loads(Path(config_file).read_text(encoding="utf-8"))

    config["project_dir"] = project_dir
    config["config_file"] = config_file

    # Validate config
    if "switchblade" not in config:
        raise MissingRequiredConfigKey("Missing [switchblade] section in config file")

    for section in REQUIRED_CONFIG_KEYS:
        if section not in config:
            raise MissingRequiredConfigKey(f"Required configuration section not found: [{section}]")
        for key in REQUIRED_CONFIG_KEYS[section]:
            if key not in config[section]:
                raise MissingRequiredConfigKey(f"Required configuration key not found: [{section}].{key}")

    return config



import os

import click

from switchbladecli.cli.config import find_config_file, get_switchblade_config
from switchbladecli.cli.lint import lint
from switchbladecli.cli.test import test
from switchbladecli.cli.update import update

ALIASES = {}


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


# We pass the project directory to all subcommands via the context
# so they can use it to find the switchblade.toml file
@click.group(cls=AliasedGroup)
@click.option("--config", "config_file", help="Path to Switchblade configuration file.")
@click.option("--verbose", is_flag=True, default=False, help="Show verbose output")
@click.pass_context
def switchbladecli(
    ctx, config_file, verbose
): 
    """Sentient Switchblade

    A part of the Beth Developer Toolbelt
    """
    config_file = find_config_file(verbose, config_file)

    project_dir = os.path.abspath(os.path.dirname(config_file))

    config = get_switchblade_config(verbose, project_dir, config_file)

    ctx.obj = {
        "config": config,
        "config_file": os.path.abspath(config_file),
        "project_dir": project_dir,
        "verbose": verbose,
    }


for command in [update, lint, test]:
    switchbladecli.add_command(command)


if __name__ == "__main__":
    switchbladecli(obj={})  # pylint: disable=no-value-for-parameter



from pathlib import Path
import click
import shutil
import subprocess  # nosec

from datetime import datetime, timezone
from mergedeep import merge
from tomlkit import dumps, loads

from switchbladecli.cli.bundle_cache import Bundle, BundleCache


class PythonPoetry:

    SKIP_UPDATE_IF_INSTALLED_WITHIN = 60*60*24  # 24 hours
    mode = "python-poetry"

    def __init__(self, config: dict, verbose: bool):
        self._config = config
        self._project_folder = Path(config["project_dir"])
        self._verbose = verbose
        self._cache = BundleCache(self._config)


    # TODO: Make this an override
    def install_dev_tools(self, bundle: Bundle):
        # Check for a virtualenv and return an error if one is not found
        check_env = subprocess.run(["poetry", "env", "info", "--path"], cwd=self._project_folder, check=False, capture_output=True)  # nosec
        if check_env.returncode != 0:
            raise click.ClickException("No virtualenv found. Please run `poetry install` first.")

        # Read pyproject.toml and remove the sections [tool.poetry.group.switchblade]
        # and [tool.poetry.group.switchblade.dependencies], and add the ones from the
        # bundle.toml file in the commit SHA folder
        self._pyproject_file = self._project_folder / "pyproject.toml"
        self._poetry_lockfile = self._project_folder / "poetry.lock"

        self._pyproject_file_raw_before = self._pyproject_file.read_text(encoding="utf-8")
        if self._poetry_lockfile.exists():
            self._poetry_lockfile_raw_before = self._poetry_lockfile.read_text(encoding="utf-8")
        else:
            self._poetry_lockfile_raw_before = None
        pyproject_config = loads(self._pyproject_file_raw_before)
        pyproject_config["tool"]["poetry"]["group"][
            "switchblade"
        ] = bundle.bundle_config["tool"]["poetry"]["group"]["switchblade"]

        # Also take all other sections beginning with 'tool.' and add them to the pyproject.toml
        # unless there is already a section with the same name in the pyproject.toml
        for section in bundle.bundle_config["tool"]:
            if not section.startswith("poetry") and section not in pyproject_config["tool"]:
                pyproject_config["tool"][section] = bundle.bundle_config["tool"][
                    section
                ]
        with open(self._pyproject_file, "w") as pyproject_toml:
            pyproject_toml.write(dumps(pyproject_config))

        # If the bundle was recently installed, we can optimistically skip the update
        latest_config = self._cache.get_latest_config()
        last_installed_on = latest_config["last_installed_on"]
        if last_installed_on:
            last_installed_on = datetime.strptime(last_installed_on, "%Y-%m-%d %H:%M:%S").replace(tzinfo=timezone.utc)
            current_datetime = datetime.now(tz=timezone.utc)
            if (current_datetime - last_installed_on).total_seconds() < self.SKIP_UPDATE_IF_INSTALLED_WITHIN:
                print("🔒 bundle dependencies were recently installed, skipping...")
                return

        subprocess.run(
            ["poetry", "update", "--only", "switchblade"],
            cwd=self._project_folder,
            check=True,
            capture_output=False,
        )

        self._cache.update_latest_config({
            "last_installed_on": datetime.now(tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")
        })


    # TODO: Place this in a superclass
    def copy_config_files(self, bundle: Bundle):
        # For each file in the bundle, check if the file exists in the project folder
        pushed_files = []
        for file in bundle.get_files():
            if (self._project_folder / file).exists():
                # If the file exists, skip it
                self._cache.log(f"SKIPPED {file}")
                continue
            else:
                # If the file doesn't exist, copy it from the commit SHA folder to the project folder
                shutil.copyfile(
                    bundle.bundle_folder / file,
                    self._project_folder / file
                )
                pushed_files.append(file)
                self._cache.log(f"PUSHED {file}")
        return pushed_files


    # TODO: Make this an override (must return False if the linter fails)
    def run_linter(self, bundle: Bundle, linter: dict):
        linter_command_list = linter["command"].split(" ")
        cmd_result = subprocess.run(
            ["poetry", "run", *linter_command_list], cwd=self._project_folder, check=False
        )
        return cmd_result.returncode == 0


    # TODO: Make this an override (must return False if the test fails)
    def run_test(self, bundle: Bundle, test: dict):
        test_command_list = test["command"].split(" ")
        cmd_result = subprocess.run(
            ["poetry", "run", *test_command_list], cwd=self._project_folder, check=False
        )
        return cmd_result.returncode == 0


    # TODO: Make this an override
    def post_cleanup(self, bundle: Bundle, pushed_files: list):
        # Restore the original pyproject.toml
        if self._pyproject_file_raw_before:
            with open(self._pyproject_file, "w") as pyproject_toml:
                pyproject_toml.write(self._pyproject_file_raw_before)
        # Restore the original poetry.lock
        if self._poetry_lockfile_raw_before:
            with open(self._poetry_lockfile, "w") as poetry_lock:
                poetry_lock.write(self._poetry_lockfile_raw_before)


    # TODO: Place this in a superclass
    def lint(self, linter_tool: str):
        # Instantiating a Bundle object will fetch the latest bundle from the source or cache
        latest_bundle = Bundle(self._config)

        bundle_config = latest_bundle.bundle_config
        bundle_mode = bundle_config["bundle"]["mode"]

        if bundle_mode != self.mode:
            click.echo(
                f"⚔️ The Switchblade bundle is for a project of mode {bundle_mode}, but this project is of mode {self.mode}.",
                err=True
            )
            raise click.Abort()

        self._cache.log(f"LINTING {latest_bundle.version}")

        # Show error message if the bundle does not contain any linters
        if "linters" not in bundle_config:
            self._cache.log(f"LINTING {latest_bundle.version} ABORTED_NO_LINTERS_IN_BUNDLE")
            click.secho(
                "⚔️ The Switchblade bundle does not contain any linters.",
                err=True, fg="red"
            )
            raise click.Abort()

        pushed_files = []
        success = False
        try:
            print("⚔️ Switchblade installing dev tools...")
            self.install_dev_tools(latest_bundle)

            print("⚔️ Switchblade copying configuration...")
            pushed_files = self.copy_config_files(latest_bundle)

            # Apply local overrides from .switchblade (merging the linter dicts) if any
            switchblade_linters_config = self._config["linters"] if "linters" in self._config else {}
            merged_linters_config = merge({}, bundle_config["linters"], switchblade_linters_config)

            success = True
            linter_names = merged_linters_config["all"] if linter_tool in ["all", None] else [linter_tool]
            for linter_name in linter_names:
                print(f"⚔️ Switchblade running {linter_name}...", end=" ")
                linter_success = self.run_linter(latest_bundle, merged_linters_config[linter_name])
                print("✅" if linter_success else "❌")
                success = success and linter_success
                self._cache.log(f"RAN {linter_name}")

        except Exception as exc:
            self._cache.log(f"LINTING {latest_bundle.version} FAILED_WITH_EXCEPTION")
            raise click.ClickException(f"Exception detected while running linters, aborting... {exc}")

        finally:
            print("⚔️ Switchblade cleaning up...")
            # Remove the files from the project folder that were added in Step 3
            for file in pushed_files:
                (self._project_folder / file).unlink()
                self._cache.log(f"POPPED {file}")

            self.post_cleanup(latest_bundle, pushed_files)

        if success:
            self._cache.log(f"LINTING {latest_bundle.version} FINISHED_SUCCESSFULLY")
            click.echo("⚔️ Linting finished successfully. 😎")
        else:
            self._cache.log(f"LINTING {latest_bundle.version} FINISHED_WITH_ERRORS")
            click.secho("⚔️ Linting finished with errors. 😭", err=True, fg="red", bold=True)

        return success


    # TODO: Place this in a superclass
    def test(self, test_tool: str):
        # Instantiating a Bundle object will fetch the latest bundle from the source or cache
        latest_bundle = Bundle(self._config)

        bundle_config = latest_bundle.bundle_config
        bundle_mode = bundle_config["bundle"]["mode"]

        if bundle_mode != self.mode:
            click.echo(
                f"⚔️ The Switchblade bundle is for a project of mode {bundle_mode}, but this project is of mode {self.mode}.",
                err=True
            )
            raise click.Abort()

        self._cache.log(f"TESTING {latest_bundle.version}")

        # Show error message if the bundle does not contain any linters
        if "tests" not in bundle_config:
            self._cache.log(f"TESTING {latest_bundle.version} ABORTED_NO_TESTS_IN_BUNDLE")
            click.secho(
                "⚔️ The Switchblade bundle does not contain any tests.",
                err=True, fg="red"
            )
            raise click.Abort()

        pushed_files = []
        success = False
        try:
            print("⚔️ Switchblade installing dev tools...")
            self.install_dev_tools(latest_bundle)

            print("⚔️ Switchblade copying configuration...")
            pushed_files = self.copy_config_files(latest_bundle)

            # Apply local overrides from .switchblade (merging the tests dicts) if any
            bundle_tests_config = bundle_config["tests"]
            switchblade_tests_config = self._config["tests"] if "tests" in self._config else {}
            merged_tests_config = merge({}, bundle_tests_config, switchblade_tests_config)

            success = True
            test_tool_names = merged_tests_config["all"] if test_tool in ["all", None] else [test_tool]
            for test_tool_name in test_tool_names:
                print(f"⚔️ Switchblade running {test_tool_name}...", end = " ")
                test_success = self.run_test(latest_bundle, merged_tests_config[test_tool_name])
                print("✅" if test_success else "❌")
                success = success and test_success
                self._cache.log(f"RAN {test_tool_name}")

        except Exception as exc:
            self._cache.log(f"TESTING {latest_bundle.version} FAILED_WITH_EXCEPTION")
            raise click.ClickException(f"Exception detected while running tests, aborting... {exc}")

        finally:
            print("⚔️ Switchblade cleaning up...")
            # Remove the files from the project folder that were added in Step 3
            for file in pushed_files:
                (self._project_folder / file).unlink()
                self._cache.log(f"POPPED {file}")

            self.post_cleanup(latest_bundle, pushed_files)

        if success:
            self._cache.log(f"TESTING {latest_bundle.version} FINISHED_SUCCESSFULLY")
            click.echo("⚔️ Tests finished successfully. 😎")
        else:
            self._cache.log(f"TESTING {latest_bundle.version} FINISHED_WITH_ERRORS")
            click.secho("⚔️ Tests finished with errors. 😭", err=True, fg="red", bold=True)

        return success



from switchbladecli.exceptions import InvalidConfigValue
from switchbladecli.modes.python_poetry import PythonPoetry

def get_tool_runner(config):
    mode_map = {
        "python-poetry": PythonPoetry
    }
    if config["switchblade"]["mode"] not in mode_map:
        raise InvalidConfigValue(f'Invalid project mode {config["switchblade"]["mode"]} specified.')

    return mode_map[config["switchblade"]["mode"]]



import click

from operator import itemgetter
from switchbladecli.modes.tool_runner import get_tool_runner


@click.command()
@click.argument("linter", required=False)
@click.pass_context
def lint(ctx, linter: str):
    """Run linter(s) on project.
    
    LINTER: Linter to run. If not specified, all linters will be run.
    """
    project_dir, config, verbose = itemgetter("project_dir", "config", "verbose")(ctx.obj)
    result = cmd_lint(verbose, project_dir, config, linter)
    if not result:
        ctx.exit(1)


def cmd_lint(verbose: bool, project_dir: str, config: dict, linter_tool: str = "all"):
    tool_runner = get_tool_runner(config)(config, verbose)
    return tool_runner.lint(linter_tool)



import click

class MissingConfiguration(click.ClickException):
    pass

class MissingRequiredConfigKey(click.ClickException):
    pass

class InvalidConfigValue(click.ClickException):
    pass

class UnrecognizedConfigKey(click.ClickException):
    pass
